from enum import Enum
from typing import List, Optional

from fakts import Fakts, get_current_fakts, Config




class GrantType(str, Enum):
    IMPLICIT = "IMPLICIT"
    PASSWORD = "PASSWORD"
    CLIENT_CREDENTIALS = "CLIENT_CREDENTIALS"
    AUTHORIZATION_CODE = "AUTHORIZATION_CODE"
    AUTHORIZATION_CODE_SERVER = "AUTHORIZATION_CODE_SERVER"

class HerreConfig(Config):
    subpath: str = ""
    secure: bool 
    host: str
    port: int
    client_id: str 
    client_secret: str
    authorization_grant_type: GrantType
    scopes: List[str]
    redirect_uri: Optional[str]
    jupyter_sync: bool = False
    username: Optional[str]
    password: Optional[str]
    timeout: int = 500

    class Config:
        group = "herre"


from abc import abstractmethod
from herre.herre import Herre
from herre.wards.variables import parse_variables
from herre import get_current_herre
import asyncio
from koil import Koil, get_current_koil
from fakts import Fakts, get_current_fakts, Config
from koil.loop import koil

class WardException(Exception):
    pass

class WardMeta(type):
    """
    
    """
    def __init__(self, name, bases, attrs):
        super(WardMeta, self).__init__(name, bases, attrs)
        if attrs["__qualname__"] != "BaseWard":
            # This gets allso called for our Baseclass which is abstract
            meta = attrs["Meta"] if "Meta" in attrs else None
            assert meta is not None, f"Please provide a Meta class in your Arnheim Model {name}"

            try:
                if meta.abstract:
                    return
            except:
                pass

            register = getattr(meta, "register", True)

            if register:
                from herre.wards.registry import get_ward_registry
                key = getattr(meta, "key", None)
                assert key is not None, f"Please provide key in your Meta class to register the Ward {attrs['__qualname__']}, or specifiy register=False"
                get_ward_registry().register_ward(key, self)



class BaseQuery(object):
    """Query
    
    A query is an instructional set of how to retrieve data from a Server through a ward. A Ward takes a query populates it with credentials through
    herre and send its to its respective instance. 

    We ship with a Standard GraphQL Query.

    This can be used as a mixin for typesaftey
    
    
    
    """


class BaseWard(metaclass=WardMeta):
    """Ward

    Wards are connectors between Models and there Corresponding Endpoints (Servers). They are automatically registered in a common ward registry
    so models can just specify the ward they want to use by referencing it in their Meta 

    Args:
        BaseModel ([type]): [description]
        metaclass ([type], optional): [description]. Defaults to ModelMeta.

    Raises:
        NotImplementedError: [description]
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """
    id: str
    configClass = Config

    def __init__(self, *args,  herre: Herre = None, koil: Koil = None, fakts: Fakts = None, max_retries = 4, **kwargs) -> None:
        self.herre = herre or get_current_herre()
        self.koil = koil or get_current_koil()
        self.fakts = fakts or get_current_fakts()
        self.connected = False
        self.transcript = None
        self.config = None
        self.max_retries = max_retries
        super().__init__()

    @abstractmethod
    async def handle_run(self, query: BaseQuery, parsed_variables: dict):
        raise NotImplementedError("Your Ward must overwrite run")

    @abstractmethod
    async def handle_connect(self):
        raise NotImplementedError("Your Ward must overwrite run")

    @abstractmethod
    async def handle_disconnect(self):
        raise NotImplementedError("Your Ward must overwrite run")


    async def negotiate(self):
        """Negotiation is a step before launching the first query to your backend service,
        it allows for initial configurations to be transfer
        """
        return None


    async def arun(self, query: BaseQuery, variables: dict = {}):
        assert isinstance(query, BaseQuery), "Query must be of type BaseQuery"
        if not self.connected:
            await self.aconnect()

        return await self.handle_run(query, await parse_variables(variables))


    def run(self, query: BaseQuery, variables: dict = {}):
        return koil(self.arun(query, variables))


    async def adisconnect(self):  
        print("Doind this here")      
        await self.handle_disconnect()
        self.transcript = None
        self.connected = False

    async def aconnect(self): 
        if not self.fakts.loaded:
            await self.fakts.aload()


        if not self.herre.logged_in:
            await self.herre.login()

        self.config = self.configClass.from_fakts(fakts=self.fakts)
        await self.handle_connect()
        self.connected = True
        self.transcript = await self.negotiate()
        
        


    class Meta:
        abstract = True








try:
    from arkitekt.packers.models.graphql import GraphQLStructure
    GraphQLModel = GraphQLStructure
except ImportError as e:
    from herre.access.model import GraphQLModel
    GraphQLModel =  GraphQLModel


from enum import Enum
from urllib.parse import quote
from aiohttp import web
import asyncio
import webbrowser
import aiohttp_cors
import uuid
def wrapped_qs_future(future):

    async def web_token_response(request):
        future.set_result(request.path_qs)
        return web.Response(text="You can close me now !")

    return web_token_response


def wrapped_post_future(future, state):

    async def web_token_response(request):

        print(state) #TODO: Implement checking for state here
        future.set_result(await request.json())
        return web.json_response(data={"ok": True})

    return web_token_response


async def wait_for_server(app, host="localhost", port="6767", timeout=1):
    try:
        await asyncio.wait_for(web._run_app(app, host="localhost", port=6767, ), timeout)
    except asyncio.TimeoutError:
        return "no token"

class RedirectError(Exception):
    pass

async def wait_for_redirect(starturl, redirect_host="localhost", redirect_port=6767, path = "/", timeout=400, print_function= False, handle_signals=False):



    webbrowser.open(starturl)

    token_future = asyncio.get_event_loop().create_future()

    app = web.Application()
    cors = aiohttp_cors.setup(app, defaults={
    "*": aiohttp_cors.ResourceOptions(
            allow_credentials=True,
            expose_headers="*",
            allow_headers="*",
        )
    })

    cors.add(app.router.add_get(path, wrapped_qs_future(token_future)), {
    "*": aiohttp_cors.ResourceOptions(
            allow_credentials=True,
            expose_headers="*",
            allow_headers="*",
        )
    })
    webserver_future = asyncio.wait_for(web._run_app(app, host=redirect_host, port=redirect_port, print=print_function,handle_signals=handle_signals), timeout)
    done, pending = await asyncio.wait([token_future, webserver_future], return_when=asyncio.FIRST_COMPLETED)

    for tf in done:
        if tf == token_future:
            redirect_qs = tf.result()
        else:
            redirect_qs = None

    for task in pending:
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

    if not redirect_qs: raise RedirectError("No redirect url provided")
    return redirect_qs


async def wait_for_post(starturl, redirect_host="localhost", redirect_port=6767, path = "/", timeout=400, print_function= False, handle_signals=False):
    

    state = uuid.uuid4()
    redirect_uri = quote(f"http://{redirect_host}:{redirect_port}{path}")

    webbrowser.open(starturl +f"?redirect_uri={redirect_uri}&state={state}")

    token_future = asyncio.get_event_loop().create_future()

    app = web.Application()
    cors = aiohttp_cors.setup(app, defaults={
    "*": aiohttp_cors.ResourceOptions(
            allow_headers="*",
        )
    })
    cors.add(app.router.add_post(path, wrapped_post_future(token_future, state)))

    print("nanan")
    webserver_future = asyncio.wait_for(web._run_app(app, host=redirect_host, port=redirect_port, print=print_function,handle_signals=handle_signals), timeout)
    done, pending = await asyncio.wait([token_future, webserver_future], return_when=asyncio.FIRST_COMPLETED)

    for tf in done:
        if tf == token_future:
            redirect_qs = tf.result()
        else:
            redirect_qs = None

    for task in pending:
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

    if not redirect_qs: raise RedirectError("No redirect url provided")
    return redirect_qs



import abc
from koil.loop import koil
from typing import Any, Dict, List

from pydantic.main import BaseModel
from herre.wards.introspection import Field, Schema, TypeKind
from herre.wards.graphql import ParsedQuery, GraphQLWard
from herre.wards.registry import get_ward_registry

class SchemaException(Exception):
    pass


class QueryResult(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getattr__(self, name: str) -> Any:
        return self[name]

    def _repr_html_(self):
        return f"<div>Query<div>"



class SchemaRegistry():

    def __init__(self) -> None:
        self.wardSchemaMap: Dict[str, Schema] = {}
        self.wardTypeModelMapMap: Dict[str, Dict[str, BaseModel]] = {}


    def register_model_for_type_and_ward(self, type: str, key: str, model: BaseModel):
        self.wardTypeModelMapMap.setdefault(key, {}).setdefault(type, model)


    async def parse_from_field(self, operation_field: Field, element, ward_key):

        if operation_field.type.kind == TypeKind.LIST:
                if operation_field.type.ofType.kind == TypeKind.OBJECT:
                    model = self.wardTypeModelMapMap[ward_key][operation_field.type.ofType.name]
                    return [model(**el) for el in element]

        elif operation_field.type.kind == TypeKind.OBJECT:
            model = self.wardTypeModelMapMap[ward_key][operation_field.type.name]
            return model(**element)

        raise SchemaException(f"No Object found for this TypeKind {operation_field.type}")


    async def expand_from_schema(self, ward_key, query, result, collapse=False):
        schema = await self.get_schema_for_ward(ward_key)

        parsed_result = {}

        for key, element in result.items():
            operation_name = query.get_operation_name_for_datakey(key)
            operation_field = [field for field in schema.queryType.fields if field.name == operation_name ][0]
            parsed_result[key] = await self.parse_from_field(operation_field, element, ward_key)

        if collapse:
            assert len(parsed_result.items()) == 1, "Cannot collapse query that has more then one operation"
            return parsed_result[query.firstchild]

        return QueryResult(**parsed_result)

        

    async def get_schema_for_ward(self, key: str):
        """Returns the schema for a ward or querys it

        Args:
            key (str): [description]
        """
        if key not in self.wardSchemaMap:
            ward = get_ward_registry().get_ward_instance(key=key)
            assert isinstance(ward, GraphQLWard), "Ward is not of instance GraphQL schemas only supported by GRAPHQL"

            introspection_query = await ward._run_async(ParsedQuery("""
                                    query Schema {
                    __schema {
                        queryType {
                        name
                        description
                        fields {
                            name
                            type {
                            kind
                            ofType {
                                kind
                                name
                                description
                            }
                            name
                            description
                            }
                            description
                            deprecationReason
                        }
                        }
                    }
                    }
            """),{})

            self.wardSchemaMap[key] = Schema(**introspection_query["__schema"])

        return self.wardSchemaMap[key]

CURRENT_SCHEMA_REGISTRY = None

def get_schema_registry():
    global CURRENT_SCHEMA_REGISTRY
    if CURRENT_SCHEMA_REGISTRY is None:
        CURRENT_SCHEMA_REGISTRY = SchemaRegistry()
    return CURRENT_SCHEMA_REGISTRY


class TypedQuery:
    """Generate a query for whatever
    """
    ward_key = None

    def __init__(self, query: str,  ward_key = None, collapse=False) -> None:
        """ A Typed Query

        Typed query takes a querystring and expands it on run from registered
        schemas if the schema is a ward GRAPHQL Sever




        Args:
            query (str): [description]
            ward_key ([type]): [description]
        """
        self.ward_key = ward_key or self.ward_key 
        assert self.ward_key, "Please Specifiy ward key"
        self.ward = get_ward_registry().get_ward_instance(self.ward_key)
        self.query = ParsedQuery(query)
        self.collapse = collapse


    async def arun(self, **variables):
        result = await self.ward._run_async(self.query, variables=variables)
        expanded = await get_schema_registry().expand_from_schema(self.ward_key, self.query, result, collapse=self.collapse)
        return expanded

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.run(**kwds)

    def run(self, **variables):
        return koil(self.arun(**variables))

    def _repr_html_(self):
        return f'<td> <th>Query for Elements {self.ward_key}</th>  <tr> {self.query.firstchild} </tr></td>'








from typing import Dict
from .base import BaseGrant
from herre.config import GrantType


class GrantRegistry:

    def __init__(self) -> None:
        self.registeredGrants: Dict[str, BaseGrant] = {}

    def register_grant(self, type: GrantType,  grant: BaseGrant):
        self.registeredGrants[type] = grant

    def get_grant_for_type(self,type):
        return self.registeredGrants[type]

    
def register_grant(type: GrantType):
    print("Registering Grant")

    def rea_decorator(grant):
        assert issubclass(grant, BaseGrant), "Grant must subclass BaseGrant"
        get_current_grant_registry().register_grant(type, grant)
        return grant

    return rea_decorator



GRANT_REGISTRY = None


def get_current_grant_registry(with_defaults=True):
    global GRANT_REGISTRY
    if not GRANT_REGISTRY:
        GRANT_REGISTRY = GrantRegistry()
        if with_defaults:
            import herre.grants.backend.app 
            import herre.grants.code_server.app


    return GRANT_REGISTRY





from abc import abstractmethod
from herre.grants.models import User

from pydantic.main import prepare_config
from herre.console.context import get_current_console
from herre.config import HerreConfig
from abc import ABC
import os 
import asyncio
import requests
import logging
import aiohttp
import shelve

from fakts import Fakts, get_current_fakts, Config

logger = logging.getLogger(__name__)


class GrantException(Exception):
    pass

class RetryException(GrantException):
    pass




class BaseGrant(ABC):
    refreshable = False
    type = None

    def __init__(self, herre_config: HerreConfig,  token_url="o/token/", authorize_url="o/authorize/", token_file: str = "token.temp", save_token = True, max_retries = 3, facts: Fakts = None, **kwargs) -> None:
        self.facts = facts or get_current_fakts()
        self.config = herre_config
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0" if self.config.secure else "1"


        self.base_url = f'{"https" if self.config.secure else "http"}://{self.config.host}:{self.config.port}{self.config.subpath}/'
        self.auth_url = self.base_url + authorize_url
        self.token_url = self.base_url + token_url
        self.userinfo_url = self.base_url + "userinfo/"
        self.refresh_url = self.token_url
        self._userinfo = None
       
        self.scopes = self.config.scopes + ["introspection"]
        self.scope = " ".join(self.scopes)
        self.token = None
        self._user = None
        self.token_file = token_file
        self.save_token = save_token
        self.max_retries = max_retries

        # State changed through Class
        self.can_refresh = self.refreshable


    @abstractmethod
    async def afetch_token(self, **kwargs):
        raise NotImplementedError()

    @property
    def user(self) -> User:
        return self._user


    @property
    def logged_in(self):
        return self.token is not None

    @property
    def access_token(self):
        assert self.token is not None, "You need to login before you can use this function. Call Herre.login()"
        return self.token["access_token"]


    async def alogin(self, force_relogin=False, retry=0, **kwargs):
        """IS SYNC!!!!
        """
        if retry > self.max_retries: raise RetryException("Grant exceeded login retries")
        self.token = None
        if not force_relogin:
            try:
                with shelve.open(self.token_file) as cfg:
                    client_id = cfg["client_id"]
                    if client_id == self.config.client_id:
                        self.token = cfg['token']
                    else:
                        logger.info("Omitting old token")

            except KeyError:
                pass
                
        print(self.token)
        print("Hallo")

        if not self.token:
            self.token = await self.afetch_token(**kwargs)

        
        assert self.token is not None, "We have received not Token back from our Grant"
        assert "access_token" in self.token, "Returned token does not have an access_token"


        print(self.token)
        print("Nano")
        if "refresh_token" in self.token:
            self.can_refresh = True

        
        try:
            async with aiohttp.ClientSession(headers={"Authorization": f"Bearer {self.access_token}"}) as session:
                async with session.get(self.userinfo_url) as resp:
                    user_json = await resp.json()
                    if "detail" in user_json:
                        raise Exception(user_json["detail"])
                        
                    try:
                        self._user = User(**user_json)
                    except:
                        self._user = None

                    if self.save_token:
                        with shelve.open(self.token_file) as cfg:
                            cfg['token'] = self.token
                            cfg['client_id'] = self.config.client_id


        except Exception as e:
            logger.error("Token Invalid")
            await self.alogin(force_relogin=True, **kwargs)
            

    async def alogout(self, force_relogin=False, **kwargs):
        """IS SYNC!!!!
        """
        try:
            with shelve.open(self.token_file) as cfg:
                cfg["token"] = None
                cfg["client_id"] = None
        except KeyError:
            pass

        self.token = None
        self._user = None


    async def arefresh(self, **kwargs):
        assert "refresh_token" in self.token, "Token had not refresh-token attached"
        async with aiohttp.ClientSession() as session:
            async with session.post(self.refresh_url, data={"grant_type": "refresh_token", "refresh_token": self.token["refresh_token"], "client_id": self.config.client_id, "client_secret": self.config.client_secret}) as resp:
                self.token =  await resp.json()
                print(self.token)
                assert "access_token" in self.token, "Returned refreshed token does not have an access_token"
                if "refresh_token" in self.token:
                    self.can_refresh = True




import logging
from rich.logging import RichHandler


def setLogging(config_path=None,**kwargs):

    logger = logging.getLogger()
    logger.setLevel("INFO")

    stream = RichHandler(markup=True)
    logging.root.setLevel("INFO")
    stream.setLevel("INFO")
    logger.addHandler(stream)








from enum import Enum
from typing import List, Optional
import fakts
from herre.config import HerreConfig
from herre.grants.code_server.app import AuthorizationCodeServerGrant
from herre.grants.backend.app import BackendGrant
from herre.grants.backend.app import BackendGrant
import os
import logging
from koil import get_current_koil, Koil
import time

from koil.loop import koil

from fakts import Fakts, get_current_fakts, Config
from herre.grants.registry import GrantRegistry, get_current_grant_registry



logger = logging.getLogger(__name__)



class HerreError(Exception):
    pass


class Herre:

    def __init__(self,
        *args,
        register = True,
        config: HerreConfig = None,
        koil: Koil = None,
        fakts: Fakts = None,
        granty_registry: GrantRegistry = None,
        **kwargs
    ) -> None:
        """ Creates A Herre Client

        Args:
            config_path (str, optional): [description]. Defaults to "bergen.yaml".
            username (str, optional): [description]. Defaults to None.
            password (str, optional): [description]. Defaults to None.
            allow_insecure (bool, optional): [description]. Defaults to False.
            in_sync (bool, optional): Should we force an in_sync modus if an event loop is already running. Loop will be send to another thread. Defaults to True.

        Raises:
            HerreError: [description]
        """
        self.config = config 
        self.fakts = fakts or get_current_fakts()
        self.grant_registry = granty_registry or get_current_grant_registry()
        self.koil = koil or get_current_koil()
        self.grant = None

        if register:
            set_current_herre(self)

        super().__init__(*args, **kwargs)




    async def alogin(self, **kwargs):
        if not self.config:
            if not self.fakts.loaded:
                await self.fakts.aload()

            self.config = HerreConfig.from_fakts(fakts=self.fakts)
        
        self.grant = self.grant_registry.get_grant_for_type(self.config.authorization_grant_type)(self.config, fakts=self.fakts)
        return await self.grant.alogin(**kwargs)

    async def alogout(self):
        assert self.grant, "We have never logged in"
        return await self.grant.alogout()

    async def arefresh(self):
        assert self.grant, "We have never logged in"
        return await self.grant.arefresh()


    def login(self, **kwargs):
        return koil(self.alogin(**kwargs))

    def logout(self):
        return koil(self.alogout())
    
    def refresh(self):
        return koil(self.arefresh())


    @property
    def logged_in(self):
        return self.grant and self.grant.logged_in

    @property
    def user(self):
        assert self.grant.logged_in, "User is not logged in"
        return self.grant.user

    @property
    def headers(self):
        assert self.grant.access_token is not None, "Access token is not set yet, please login?"
        return {"Authorization": f"Bearer {self.grant.access_token}"}




CURRENT_HERRE = None

def get_current_herre(**kwargs):
    global CURRENT_HERRE
    if not CURRENT_HERRE:
        CURRENT_HERRE = Herre(**kwargs)
    return CURRENT_HERRE

def set_current_herre(herre):
    global CURRENT_HERRE
    CURRENT_HERRE = herre


from herre.config import GrantType
from herre.grants.base import BaseGrant
from oauthlib.oauth2 import WebApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session
from aiohttp import ClientSession
import requests
import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer
import aiohttp
from aiohttp import web
import asyncio
import logging

from herre.grants.registry import register_grant


logger = logging.getLogger(__name__)







def wrapped_future(future):
    async def web_token_response(request):
        future.set_result(request.path_qs)
        return web.Response(text="You can close me now !")

    return web_token_response


async def wait_for_server(app, host="localhost", port="6767", timeout=1):
    try:
        await asyncio.wait_for(web._run_app(app, host="localhost", port=6767, ), timeout)
    except asyncio.TimeoutError:
        return "no token"



@register_grant(GrantType.AUTHORIZATION_CODE)
@register_grant(GrantType.AUTHORIZATION_CODE_SERVER)
class AuthorizationCodeServerGrant(BaseGrant):
    refreshable = True

    async def afetch_token(self, **kwargs):

        self.web_app_client = WebApplicationClient(self.config.client_id, scope=self.scope)

        # Create an OAuth2 session for the OSF
        self.session = OAuth2Session(
            self.config.client_id, 
            self.web_app_client,
            scope=self.scope, 
            redirect_uri=self.config.redirect_uri,
            
        )

        auth_url, state = self.session.authorization_url(self.auth_url)

        webbrowser.open(auth_url)

        token_future = asyncio.get_event_loop().create_future()

        app = web.Application()
        app.router.add_get("/", wrapped_future(token_future))
        webserver_future = asyncio.wait_for(web._run_app(app, host="localhost", port=6767, print=lambda x: logger.info(x),handle_signals=False), self.config.timeout)
        done, pending = await asyncio.wait([token_future, webserver_future], return_when=asyncio.FIRST_COMPLETED)

        for tf in done:
            if tf == token_future:
                path = tf.result()
                print(path)
            else:
                path = None

        for task in pending:
            task.cancel()

            try:
                await task
            except asyncio.CancelledError:
                pass

        if path:
            return self.session.fetch_token(self.token_url, client_secret=self.config.client_secret, authorization_response=path, state=state)

        return None



from herre import Herre
from qtpy import QtCore, QtWidgets

class QtHerre(Herre, QtWidgets.QWidget):
    login_signal = QtCore.Signal(bool)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args,  **kwargs)

    
    async def alogin(self, **kwargs):
        nana = await super().alogin(**kwargs)
        self.login_signal.emit(True)
        return nana


    async def alogout(self, **kwargs):
        nana = await super().alogout(**kwargs)
        self.login_signal.emit(False)
        return nana


import asyncio
from typing import Dict, Type
from herre import get_current_herre
from herre.wards.base import BaseWard


class NoWardRegistered(Exception):
    pass



class WardRegistry:

    def __init__(self) -> None:
        self.keyWardClassMap: Dict[str, Type[BaseWard]] = {}
        self.keyInstanceMap: Dict[str, BaseWard] = {}

    def register_ward(self, identifier, wardClass: Type[BaseWard]):
        assert identifier not in self.keyWardClassMap, "We cannot register another Ward for this key"
        self.keyWardClassMap[identifier] = wardClass

    def get_ward(self, identifier) -> BaseWard:
        try:
            return self.keyWardClassMap[identifier]
        except KeyError as e:
            raise NoWardRegistered(f"No Structure registered for identifier {identifier}") from e

    def get_ward_instance(self, key) -> BaseWard:
        if key in self.keyInstanceMap: return self.keyInstanceMap[key]
        self.keyInstanceMap[key] = self.keyWardClassMap[key]()
        return self.keyInstanceMap[key]


    # Handling Connection

    async def connect_all(self):
        unconnected_wards = [ward for key, ward in self.keyInstanceMap.items() if not ward.connected]
        await asyncio.gather(*[ward._connect() for ward in unconnected_wards]) 

    async def unconnect_all(self):
        connected_wards = [ward for key, ward in self.keyInstanceMap.items() if ward.connected]
        await asyncio.gather(*[ward._disconnect() for ward in connected_wards])





WARD_REGISTRY = None

def get_ward_registry():
    global WARD_REGISTRY
    if not WARD_REGISTRY:
        WARD_REGISTRY = WardRegistry()
    return WARD_REGISTRY